%UNTRAMPC Tracking Predictive Control of unstable plant. (Script file)
%
% Implements unconstrained MPC for unstable SISO discrete-time system 
% defined by z-transform transfer function, with tracking of constant 
% set-point. As TRACKMPC, but `realigns'model on latest plant outputs.
%
% The following are editable parameters (most have defaults):
% Tref: Desired closed-loop response time
% Ts:   Sampling interval 
% nump, denp, plant: Plant definition
% numm, denm, model: Model definition
% P: Vector of coincidence points
% M: Control horizon
% tend: Duration of simulation
% setpoint: Setpoint trajectory (column vector) - length must exceed
%           no of steps in simulation by at least max(P).
% umpast, uppast, ympast, yppast: Initial conditions of plant & model.
%
% NOT DEBUGGED YET
%
% Assumes Matlab 5.3. Uses Control Toolbox LTI object class. 
% No other toolboxes required.

%% J.M.Maciejowski, 9 March 1999. Revised 15.12.99.
%% Copyright(C) 1999. All Rights Reserved.
%% Cambridge University Engineering Department.

%%%%%%%%%%%%%%%%% PROBLEM DEFINITION:

% Define time-constant of reference trajectory Tref:
Tref = 0;

% Define sampling interval Ts (default Tref/10):
if Tref == 0,
  Ts = 1;
else
  Ts = Tref/10;
end

% Define plant as SISO discrete-time 'lti' object 'plant' in transfer
% function form: 
nump=1;
denp=[1,-1.6,0.55];
plant = tf(nump,denp,Ts);
nnump = length(nump)-1; % Degree of plant numerator
ndenp = length(denp)-1; % Degree of plant denominator
if ndenp-nnump<1, error('Plant must be strictly proper'), end;
if any(abs(roots(denp))>1), disp('Warning: Unstable plant'), end

% Define model as SISO discrete-time 'lti' object 'model' 
% (default model=plant):
numm=1;
denm=[1,-1.6,0.56];
model = tf(numm,denm,Ts);
nnumm = length(numm)-1; % Degree of model numerator
ndenm = length(denm)-1; % Degree of model denominator
if ndenm-nnumm<1, error('Model must be strictly proper'), end;

nump=[zeros(1,ndenp-nnump-1),nump]; % Pad numerator with leading zeros
numm=[zeros(1,ndenm-nnumm-1),numm]; % Pad numerator with leading zeros

% Define prediction horizon P (default 0.8*Tref):
if Tref == 0,
  P = 5;
else
  P = round(0.8*Tref/Ts);
end
% Define corresponding vector of weights on errors:
ywt = [zeros(P-1,1);1]; % Zero weight on first P-1 predicted errors, 
                        % then 1 on P'th predicted error.

% Define control horizon (default 1):
M = 1;
% Define corresponding vector of control penalty weights (default 0):
uwt = 0;

% Compute model step response values over coincidence horizon:
stepresp = step(model,max(P));
respmat = [];
for i=0:M-1,
  respmat = [respmat,stepresp(P-i)]; % Hankel-like response matrix
end

% Compute reference error factor at coincidence points:
% (Exponential approach of reference trajectory to set-point)
if Tref == 0,  % Immediate jump back to set-point trajectory
  errfac = zeros(length(P),1);
else
  errfac = exp(-P*Ts/Tref);
end

%%%%%%%%%%%%%%%% SIMULATION PARAMETERS:

if Tref == 0,
  tend = 100*Ts;
else
  tend = 10*Tref; % Duration of simulation (default 10*Tref)
end
nsteps = floor(tend/Ts); % (Number of steps in simulation).
tvec = (0:nsteps-1)'*Ts; % Column vector of time points (first one 0)

% Define set-point (column) vector (default constant 1):
setpoint = ones(nsteps+max(P),1);

% Define vectors to hold input and output signals, initialised to 0:
uu = zeros(nsteps,1); % Input
yp = zeros(nsteps,1); % Plant Output
ym = zeros(nsteps,1); % Model Output

% Initial conditions:
umpast = zeros(ndenm,1);
uppast = zeros(ndenp,1);
ympast = zeros(ndenm,1);  % For model response
yppast = zeros(ndenp,1);  % For plant response

%%%%%%%%%%%%%%%% SIMULATION:

for k=1:nsteps,

  % Define reference trajectory at coincidence points:
  errornow = setpoint(k)-yp(k);
  reftraj = setpoint(k+P) - errornow*errfac;

  % Free response of model over prediction horizon:
  yfpast = ympast;
  ufpast = umpast;
  for kk=1:max(P), % Prediction horizon
    ymfree(kk) = numm*ufpast-denm(2:ndenm+1)*yfpast;
    yfpast=[ymfree(kk);yfpast(1:length(yfpast)-1)];
	ufpast=[ufpast(1);ufpast(1:length(ufpast)-1)];
  end
  
  % Compute input signal uu(k):
  referr = reftraj - ymfree(P); % reference traj - free response at
                                % coincidence points
  %%%%% NEXT LINE GIVES TRACKING PROPERTY:
  %%%%% but this does not work with `realigned' model
  referr = referr - (yppast(1)-ympast(1)); % Assume latest plant-model
                                           % difference will persist.
  %%%%%
  dutraj = respmat\referr; % No weights yet. This is over control horizon
  if k>1,
    uu(k) = dutraj(1)+uu(k-1); % Apply first one only.
  else
    uu(k) = dutraj(1)+umpast(1);
  end

  % Simulate plant and model:
  uppast = [uu(k);uppast(1:length(uppast)-1)];    % Update past plant inputs
  yp(k+1) = -denp(2:ndenp+1)*yppast+nump*uppast;
  yppast = [yp(k+1);yppast(1:length(yppast)-1)];  % Update past plant outputs


  % Simulate model:
  umpast = [uu(k);umpast(1:length(umpast)-1)];   % Update past model inputs
  ym(k+1) = -denm(2:ndenm+1)*ympast+numm*umpast;
  ympast = yppast; % Update past model outputs - realign model on plant.

end % of simulation

%%%%%%%%%%%%%%%% PRESENTATION OF RESULTS:

figure
subplot(211)
% Plot output, solid line and set-point, dashed line:
plot(tvec,yp(1:nsteps),'-',tvec,setpoint(1:nsteps),'--');
grid; title('Plant output (solid) and set-point (dashed)')
xlabel('Time')

subplot(212)
% plot input signal as staircase graph:
stairs(tvec,uu,'-');
grid; title('Input')
xlabel('Time')
